// -*-C++-*-
//
//  This file is part of the Siena Simulator, a simple discrete-event
//  simulator used with the Siena project.  See
//  http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2001 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: ssim.h,v 1.11 2003/06/28 20:09:02 carzanig Exp $
//
#ifndef _ssim_h
#define _ssim_h

#include <vector>
#include <map>

using std::multimap;
using std::map;
using std::vector;

/** \mainpage SSim C++ API Documentation
 *
 *  This documentation describes the C++ application programming
 *  interface to <a
 *  href="http://www.cs.colorado.edu/~carzanig/ssim/">Ssim</a>, a very
 *  simple discrete-event sequential simulation library.  The
 *  simulator implemented by <em>ssim</em> executes reactive
 *  processes.  These processes are defined in terms of discrete
 *  execution steps performed at given times, or in response to an
 *  interaction (an <em>event</em>) with another process.  During the
 *  execution of an action, a process may schedule other future
 *  actions for itself, or it may signal events to another processes,
 *  which will respond by processing the signalled events at the given
 *  time.  The simulation terminates when no more actions are
 *  scheduled.
 *  
 *  <p>The ssim library consists of essentially two classes defined
 *  within the <code>ssim</code> namespace: \link ssim::Sim\endlink
 *  which defines the interface to the simulator, and \link
 *  ssim::Process\endlink which defines the interface for processes
 *  that the simulator can execute.  Sim offers basic primitives for
 *  scheduling actions, signaling \link ssim::Event events\endlink,
 *  and for creating, starting, and stopping processes.  Process
 *  declares the execution steps scheduled when a process is \link
 *  ssim::Process::init() started\endlink, \link
 *  ssim::Process::process_timeout() periodically scheduled\endlink,
 *  \link ssim::Process::process_event() signaled\endlink, and \link
 *  ssim::Process::stop() stopped\endlink.
 *
 *  <p>The execution of the simulation is based on a <em>virtual
 *  clock</em> that represents the time in the simulated world.  The
 *  virtual clock is simply a counter, therefore the time unit is
 *  determined by the semantics of the simulated processes. The
 *  initial value of the virtual clock is 0.  The passage of (virtual)
 *  time in the simulated world is explicitly controlled by each
 *  process, essentially in three ways: <ul>
 *
 *  <li>by "sleeping".  That is, by scheduling an event for itself
 *  after a given timeout (see \link ssim::Sim::set_timeout()
 *  set_timeout()\endlink);
 *
 *  <li>by signaling events to other processes with a given delay (see
 *  \link ssim::Sim::signal_event(const Event*,ProcessId,Time)
 *  signal_event()\endlink);
 *
 *  <li>by explicitly declaring the duration of an action or an
 *  execution step, thereby advancing their clock (see \link
 *  ssim::Sim::advance_delay(Time) advance_delay()\endlink.)
 *
 *  </ul>
 *
 *  <p>In addition to the basic Process class, the library provides a
 *  utility class \link ssim::ProcessWithPId ProcessWithPId\endlink,
 *  that automate some common procedures for process implementations.
 **/

/** \example twoprocesses.cc
 * 
 *  This is an example of how to program a simple simulation with two
 *  very simple iterative processes.  To compile this example on a
 *  GNU/Linux system or on other Unix-like systems, assuming SSim has
 *  been installed in $prefix, execute:
 * 
 *  <pre>
 *  c++ twoprocesses.cc -o twoprocesses -I$prefix/include -L$prefix/lib -lssim
 *  </pre>
 **/

/** \example bs.cc
 * 
 *  This extensive example implements a simulation of a simple
 *  supply-chain system with a manufacturer, a retailer, and a number
 *  of buyers.  The example consists of a header file and an
 *  implementation file.  <p>Thanks to Matt Rutherford for creating
 *  this example.
 *
 *  \include bs.h
 **/

/** \file ssim.h 
 *
 *  This header file defines the simulator API.
 **/

/** \namespace ssim
 *
 *  @brief name space for the Siena simulator.
 *
 *  This namespace groups all the types and functionalities associated
 *  with the Siena simulator.  These include:
 *
 *  <ol>
 *  <li>the simulator API
 *
 *  <li>the base classes for processes and events
 *
 *  <li>a few other utility classes
 *  </ol>
 **/
namespace ssim {

/** @brief version identifier for this ssim library
 **/
extern const char *		Version;

/** @brief process identifier type
 **/
typedef unsigned long		ProcessId;

/** @brief no process will be identified by NULL_PROCESSID
 **/
const ProcessId			NULL_PROCESSID = 0;

/** @brief virtual time type
 *
 *  This type represents the basic time in the virtual (simulated)
 *  world.  The actual semantics of these time units is determined by
 *  the simulation implementation.
 *
 *  @see Sim::advance_delay(Time), 
 *       Sim::set_timeout(Time),
 *       Sim::signal_event(const Event*,ProcessId, Time).
 **/
typedef unsigned long		Time;

/** @brief beginning of time
 **/
const Time			INIT_TIME = 0;

/** @brief event-specifier type
 *
 *  Event types provide a way to partition the space of events and
 *  signals so that processes can respond with the appropriate
 *  actions.
 **/
typedef unsigned		EventType;

/** @brief basic event in the simulation.
 *
 *  This base class represents a piece of information or a signal
 *  exchanged between two processes through the simulator.
 * 
 *  @see Sim::signal_event(const Event*,ProcessId),
 *       Sim::signal_event(const Event*,ProcessId, Time),
 *       Process::process_event(const Event*).
 **/
class Event {
 public:
    /** @brief type of this event
     **/
    EventType			type;

				Event();
				Event(EventType);
    virtual			~Event();

 private:
    mutable unsigned refcount;
    friend class Sim;
};

/** @brief Virtual class (interface) representing processes running
 *  within the simulator.
 *
 *  A simulated process must implement this basic interface.
 **/
class Process {
 public:
    virtual			~Process();

    /** @brief action when the process is created.
     *
     *  This execution step is scheduled when the process is created
     *  through Sim::create_process.  This action is executed
     *  before any event or timeout.
     **/
    virtual void		init(void);

    /** action executed in response to an event signaled to this process.
     *
     *  Notice that the signaled event should not be used outside this
     *  method, other than by signaling it to a process through
     *  Sim::signal_event.  Specifically, the Event pointer passed to
     *  this method should be considered invalid outside this method.
     *  in fact, the simulator proceeds to de-allocate every event
     *  object that is not signaled to any other process.
     *
     *  <p>The implementation of this method may specify the duration
     *  of the actions associated with this response using the \link
     *  Sim::advance_delay(Time) advance_delay method\endlink.  By
     *  default, the duration of an action is 0.
     *
     *  @see Sim::signal_event(const Event*,ProcessId),
     *       Sim::signal_event(const Event*,ProcessId,Time)
     *       Sim::advance_delay(Time).
     **/
    virtual void		process_event(const Event * msg);

    /** action executed in response to a timeout.
     *
     *  This method defines the actions explicitly scheduled for this
     *  process by the process itself.  These actions are scheduled by
     *  calling Sim::set_timeout(Time).
     *
     *  <p>The implementation of this method may specify the duration
     *  of the actions associated with this response using the \link
     *  Sim::advance_delay(Time) advance_delay method\endlink.  By
     *  default, the duration of an action is 0.
     **/
    virtual void		process_timeout();

    /** @brief executed when the process is explicitly stopped.
     *
     *  A process is stopped by a call to
     *  Sim::stop_process(ProcessId).  This method is executed
     *  immediately after the process has processed all the events or
     *  timeouts scheduled before the call to Sim::stop_process.
     **/
    virtual void		stop(void);
};

/** @brief utility Process class providing a utility interface with the
 *  simulator.
 *
 *  This is a sligtly more advanced Process class that provides 
 *  automatic management of the process-id.
 **/
class ProcessWithPId : public Process {
 public:
    /** @brief activates this process within the simulator.
     *
     *  Creates a simulator process with this process object.  The
     *  same ProcessWithPId can be activated only once.
     *
     *  @return the ProcessId of the created simulator process. Or
     *  NULL_PROCESSID if this process object is already associated
     *  with a simulation process.
     *
     *  @see Sim::create_process(Process*,char)
     **/
    ProcessId			activate(char mode = 0);

    /** @brief process id of this process.
     *
     *  @return the id of the simulation process with this objectk,
     *  or NULL_PROCESSID if no process is associated with this
     *  object.
     **/
    ProcessId			pid() const;

    ProcessWithPId();

 private:
    ProcessId process_id;
};

/** @brief a generic discrete-event sequential simulator
 *
 *  This class implements a generic discrete-event sequential
 *  simulator.  Sim maintains and executes a time-ordered schedule of
 *  actions (or discrete events).
 **/ 
class Sim {
public:
    /** @brief creates a new process
     *
     *  creates a new process with the given Process object.
     *  <code>mode</code> specifies the execution flags of the
     *  process.  If the <code>P_SEQUENTIAL</code> flag is set, then
     *  the process will process one event at a time with respect to
     *  the simulation virtual time, otherwise (default) the actions
     *  of this process will be considered reentrant and may be
     *  executed in parallel.  By default, the simulator discards the
     *  events signalled to a <code>P_SEQUENTIAL</code> process while
     *  that process is busy executing other actions.  The
     *  <code>P_QUEUEING</code> flag can be set to instruct the
     *  simulator to queue those events.  In this case, the simulator
     *  will deliver signals to that process as the process is
     *  available to respond to them.
     **/
    static ProcessId		create_process(Process *, char mode = 0);

    /** @brief <code>P_SEQUENTIAL</code> flag */
    static const char		P_SEQUENTIAL = 0x02;	
    /** @brief <code>P_QUEUEING</code> flag */
    static const char		P_QUEUING = 0x04; 

    /** @brief stops the execution of a given process */
    static int			stop_process(ProcessId);
    /** @brief stops the execution of the current process */
    static void			stop_process();

   /** @brief clears out internal data structures
    *
    *  Resets the simulator making it available for a completely new
    *  simulation.  All scheduled actions are deleted together with
    *  the associated events.  All process identifiers returned by
    *  previoius invocations of \link create_process(Process*,char)
    *  create_process\endlink are invalidated by this operation.
    *  Notice however that it is the responsibility of the simulation
    *  programmer to delete process objects used in the simulation.
    **/
    static void                 clear();

    /** @brief signal an event to the given process 
     *
     *  Signal an event to the given process.  The response is
     *  scheduled for the current time.
     *
     *  @see Process::process_event(const Event * msg);
     **/
    static int			signal_event(const Event *, ProcessId);
    /** @brief signal an event to the given process at the given time 
     *
     *  @see Process::process_event(const Event * msg);
     **/
    static int			signal_event(const Event *, ProcessId, Time);

    /** @brief sets a timeout for the current process
     *
     *  Schedules the execution of process_timeout() on the current
     *  process after the given amount of (virtual) time.
     **/
    static void			set_timeout(Time);

    /** @brief advance the execution time of the current process.
     *
     *  This method can be used to specify the duration of certain
     *  actions, or certain steps within the same action.  For example:
     *
     *  \code
     *  class SomeProcess : public Process {
     *      //...
     *      virtual void process_event(const Event * e) {
     *          //
     *          // suppose this response is called at (virtual) time X
     *
     *          // ...do something here...
     *
     *          // the above actions have a default duration of 0,
     *          // therefore the following event is scheduled at time X + 5
     *          //
     *          signal_event(e, p, 5);
     *     
     *          advance_delay(10);
     *
     *          // ...do something else here...
     *
     *          // advance_delay(10) specifies a duration of 10, therefore 
     *          // the following timeout is scheduled at time X + 15;
     *          //
     *          set_timeout(5);
     *      }
     *  };
     *  \endcode
     *
     *  @see Sim::clock()
     **/
    static void			advance_delay(Time);
    
    /** @brief returns the current process
     *
     *  Example:
     *  \code
     *  void SimpleProcess::process_event(const Event *e) {
     *      cout << "my process id is: " << Sim::this_pocess() << endl;
     *  }
     *  \endcode
     **/
    static ProcessId		this_process();

    /** @brief returns the current virtual time for the current process
     *  
     *  Example:
     *  \code
     *  void LoopProcess::process_timeout() {
     *      cout << "Here I am at time:" << Sim::clock() << endl;
     *      set_timeout(20)
     *      cout << "I just scheduled myself again for time: " 
     *           << Sim::clock() + 20 << endl;
     *  }
     *  \endcode
     *
     *  @see advance_delay(Time)
     **/
    static Time			clock();
    
    /** @brief starts execution of the simulation */
    static void			run_simulation();
    /** @brief stops execution of the simulation */
    static void			stop_simulation();

private:
    /** @brief virtual time 
     *
     *  time in the simulated world.
     **/
    static Time			current_time;
    /** currently running process **/ 
    static ProcessId		current_process;
    /** duration of the current action **/
    static Time			current_delay;
    /** is simulation running? **/
    static bool			running;

    enum ActionType { A_Event, A_Timeout, A_Init, A_Stop };
    
    struct Action {
	ActionType	type;
	ProcessId	pid;
	const Event *	event;
	
			Action(ActionType, ProcessId);
			Action(ActionType, ProcessId, const Event *);

			Action(const Action &);
	Action &	operator = (const Action &);
    };

    typedef multimap<Time, Action>	ActionsTable;
    static ActionsTable			actions;

    static void		schedule(const Action &, Time);
    static void		schedule_now(const Action &);

    //
    // internal process status (within Sim)
    // 
    static const char P_TERMINATED	= 0x01;	// running(0)/terminated(1) 

    struct PDescr {
	Process * 	process;
	unsigned char	status;
	Time		total_action_time;
	Time		available_at;

	PDescr();
	PDescr(Process * p, char mode);
    };

    typedef vector<PDescr>		PsTable;
    static PsTable			processes;
};

//
// inline implementations
//
inline Event::Event(): refcount(0) {};
inline Event::Event(EventType t): type(t), refcount(0) {};
inline Event::~Event() {};

inline Process::~Process() {};

inline void Process::init(void) {};
inline void Process::stop(void) {};
inline void Process::process_event(const Event *) {};
inline void Process::process_timeout() {};

inline Sim::PDescr::PDescr(): 
    process((Process*)0), status(P_TERMINATED), 
    total_action_time(INIT_TIME), available_at(INIT_TIME) {};

inline Sim::PDescr::PDescr(Process * p, char mode): 
    process(p), status(mode), 
    total_action_time(INIT_TIME), available_at(INIT_TIME) {};

inline Sim::Action::Action(ActionType i, ProcessId p, const Event *e):
    type(i), pid(p), event(e) {}

inline Sim::Action::Action(ActionType i, ProcessId p):
    type(i), pid(p), event((Event*)0) {}

inline Sim::Action::Action(const Action &a):
    type(a.type), pid(a.pid), event(a.event) {}

inline Sim::Action & Sim::Action::operator = (const Action &a) {
    type = a.type; pid = a.pid; event = a.event;
    return *this;
}

inline void Sim::set_timeout(Time t) {
    schedule(Action(A_Timeout, current_process), t); 
}

inline void Sim::stop_process() {
    schedule_now(Action(A_Stop, current_process)); 
}

inline int Sim::stop_process(ProcessId pid) {
    if (processes[pid].status & P_TERMINATED) return -1;
    schedule_now(Action(A_Stop, pid)); 
    return 0;
}

inline void Sim::stop_simulation() {
    running = false;
}

inline void Sim::advance_delay(Time delay) {
    if (!running) return;
    processes[current_process].total_action_time += delay;
    current_delay += delay;
}

inline ProcessId Sim::this_process() {
    return current_process;
}

inline Time Sim::clock() {
    return current_time + current_delay;
}

inline int Sim::signal_event(const Event * e, ProcessId pid) {
    if (processes[pid].status & P_TERMINATED) return -1;
    schedule_now(Action(A_Event, pid, e));
    e->refcount++;
    return 0;
}

inline int Sim::signal_event(const Event * e, ProcessId pid, Time delay) {
    if (processes[pid].status & P_TERMINATED) return -1;
    schedule(Action(A_Event, pid, e), delay);
    e->refcount++;
    return 0;
}

inline void Sim::schedule(const Action & action, Time t) {
    actions.insert(ActionsTable::value_type(clock() + t, action));
}

inline void Sim::schedule_now(const Action & action) {
    actions.insert(actions.begin(), ActionsTable::value_type(clock(), action));
}

inline ProcessId ProcessWithPId::activate(char mode) {
    if (process_id == NULL_PROCESSID) {
	return process_id = Sim::create_process(this, mode);
    } else {
	return NULL_PROCESSID;
    }
}

inline ProcessWithPId::ProcessWithPId(): process_id(NULL_PROCESSID) {};

inline ProcessId ProcessWithPId::pid() const {
	return process_id;
}

}; // end namespace ssim

#endif /* _ssim_h */

